/*****************************************************************************
 *                     Yumetech, Inc Copyright (c) 2004 - 2006
 *                               Java Source
 *
 * This source is licensed under the GNU LGPL v2.1
 * Please read http://www.gnu.org/copyleft/lgpl.html for more information
 *
 * This software comes with the standard NO WARRANTY disclaimer for any
 * purpose. Use it at your own risk. If there's a problem you get to fix it.
 *
 ****************************************************************************/

package org.j3d.aviatrix3d.output.graphics;

// External imports
import javax.media.opengl.*;

import java.util.HashMap;
import java.util.Map;

import javax.vecmath.Matrix4f;
import javax.vecmath.Point3f;

// Local imports
import org.j3d.aviatrix3d.PBufferTextureSource;

import org.j3d.aviatrix3d.pipeline.graphics.GraphicsInstructions;
import org.j3d.aviatrix3d.pipeline.graphics.GraphicsOutputDevice;
import org.j3d.aviatrix3d.rendering.DeletableRenderable;
import org.j3d.aviatrix3d.rendering.ShaderSourceRenderable;

import org.j3d.util.DefaultErrorReporter;
import org.j3d.util.ErrorReporter;

/**
 * Implementation of the most basic drawable surface, supporting the minimal
 * number of features that is to be used by other, more complete
 * implementations.
 * <p>
 *
 * This implementation of GraphicsOutputDevice renders to a normal GLCanvas
 * or GLJPanel (depending on if "lightweight" is true) instance
 * and provides pBuffer support as needed. Stereo support is not provided and
 * all associated methods always indicate negative returns on query about
 * support.
 *
 * @author Justin Couch
 * @version $Revision: 3.20 $
 */
public abstract class BaseSurface
    implements GraphicsOutputDevice
{
    /** Message when the GL context failed to initialise */
    protected static final String FAILED_CONTEXT_MSG =
        "The GL context failed to initialise for some mysterious reason";

    /** The initial size of the children list */
    private static final int LIST_START_SIZE = 20;

    /** The real canvas that we draw to */
    protected GLDrawable canvas;

    /** The context of the main canvas */
    protected GLContext canvasContext;

    /** Abstract representation of the underlying primary renderer */
    protected RenderingProcessor canvasRenderer;

    /** Local storage of the nodes that need to be rendered */
    protected GLPbuffer[] renderableList;

    /** Number of items in the renderable list */
    protected int numRenderables;

    /** Flag to say that init has been called on the canvas */
    protected boolean initComplete;

    /** List of extensions we need to check on at startup */
    private String[] extensionList;

    /** Number of extension types to check for */
    private int numExtensions;

    /** Maps from a pBuffer instance to the rendererProcessor for it */
    private Map rendererMap;

    /** Maps an offscreen texture to it's GLAutoDrawable representation */
    private Map glDrawableMap;

    /** Used to fetch clear colour values when setting up pBuffers */
    private float[] colourTmp;

    /** Flag indicating if this surface is shared with another */
    protected BaseSurface sharedSurface;

    /** Flag to say whether the underlying factory can create pbuffers */
    protected boolean canCreatePBuffers;

    /** Error reporter used to send out messages */
    protected ErrorReporter errorReporter;

    /** Flag indicating whether rendering should be stopped right now */
    protected boolean terminate;

    /** Single threaded rendering mode operation state. Defaults to false. */
    protected boolean singleThreaded;

    /**
     * Construct a surface shares it's GL context with the given surface. This
     * is useful for constructing multiple view displays of the same scene graph,
     * but from different viewing directions, such as in a CAD application.
     * <p>
     * If the sharedWith parameter is null, then this is just treated as an
     * ordinary non-shared frame. The return flag will be set appropriately.
     *
     * @param sharedWith The surface that you'd like this surface to share
     *    the GL context with, if possible. May be null.
     */
    protected BaseSurface(BaseSurface sharedWith)
    {
        sharedSurface = sharedWith;
        terminate = false;
        singleThreaded = false;
    }

    //---------------------------------------------------------------
    // Methods defined by GraphicsOutputDevice
    //---------------------------------------------------------------

    /**
     * Check to see whether this surface supports stereo rendering - which is
     * does not. Always returns false. May be overridden by derived class to
     * provide a different answer.
     *
     * @return false Stereo is not available
     */
    public boolean isStereoAvailable()
    {
        return false;
    }

    /**
     * Check to see whether this surface supports Quad buffer stereo rendering
     * - which it does not. Always returns false for this implementation. May
     * be overridden by derived class to provide a different answer.
     *
     * @return false The surface does not support stereo at all
     */
    public boolean isQuadStereoAvailable()
    {
        return false;
    }

    /**
     * Set the eye separation value when rendering stereo. The default value is
     * 0.33 for most applications. The absolute value of the separation is
     * always used. Ignored for this implementation. May be overridden by
     * derived class to provide a different answer.
     *
     * @param sep The amount of eye separation
     */
    public void setStereoEyeSeparation(float sep)
    {
    }

    /**
     * Get the current eye separation value - always returns 0. May be
     * overridden by derived class to provide a different answer.
     *
     * @return sep The amount of eye separation
     */
    public float getStereoEyeSeparation()
    {
        return 0;
    }

    /**
     * Set the rendering policy used when handling stereo. The policy must be
     * one of the _STEREO constants defined in this interface. May be
     * overridden by derived class to provide a different answer.
     *
     * @param policy The policy to currently use
     * @throws IllegalArgumentException The policy type is not one of the legal
     *    selections.
     */
    public void setStereoRenderingPolicy(int policy)
    {
    }

    /**
     * Get the current stereo rendering policy in use. If not explicitly set by
     * the user, then it will default to <code>NO_STEREO</code>. May be
     * overridden by derived class to provide a different answer.
     *
     * @return One of the *_STEREO values
     */
    public int getStereoRenderingPolicy()
    {
        return NO_STEREO;
    }

    /**
     * Queue up a collection of shader objects for processing on the next
     * frame. These processing requests are for either initialisation or log
     * handling, not for rendering.
     *
     * @param initList The shaders needing initialisation
     * @param numInit The number of shaders needing initialisation
     * @param logList The shaders needing log fetching
     * @param numLog The number of shaders needing log fetching
     */
    public void queueShaderRequests(ShaderSourceRenderable[] initList,
                                    int numInit,
                                    ShaderSourceRenderable[] logList,
                                    int numLog)
    {
        canvasRenderer.queueShaderRequests(initList, numInit, logList, numLog);

        // Take local reference in case the setDrawableObjects decides to
        // update values right now.
        int count = numRenderables;
        GLPbuffer[] surfaces = renderableList;

        if(count > 1)
        {
            for(int i = 0; i < count; i++)
            {
                if(surfaces[i] != null)
                {
                    RenderingProcessor rp =
                        (RenderingProcessor)rendererMap.get(surfaces[i]);
                    rp.queueShaderRequests(initList, numInit, logList, numLog);
                }
            }
        }
    }

    /**
     * Set the background colour that this surface should be cleared to before
     * the drawing step. Colours range from 0 to 1 in the normal manner.
     *
     * @param r The red component of the background clear colour
     * @param g The green component of the background clear colour
     * @param b The blue component of the background clear colour
     * @param a The alpha component of the background clear colour
     */
    public void setClearColor(float r, float g, float b, float a)
    {
        canvasRenderer.setClearColor(r, g, b, a);
    }

    /**
     * Set whether we should always force a local colour clear before
     * beginning any drawing. If this is set to false, then we can assume that
     * there is at least one background floating around that we can use to
     * clear whatever was drawn in the previous frame, and so we can ignore the
     * glClear(GL.GL_COLOR_BUFFER_BIT) call. The default is set to true.
     *
     * @param state true if we should always locally clear first
     */
    public void setColorClearNeeded(boolean state)
    {
        canvasRenderer.setColorClearNeeded(state);
    }

    /**
     * Update the list of items to be rendered to the current list. Draw them
     * at the next oppourtunity.
     *
     * @param commands The list of drawable surfaces to render
     * @param numValid The number of valid items in the array
     */
    public void setDrawableObjects(GraphicsInstructions[] commands,
                                   int numValid)
    {
        if(renderableList.length < numValid)
            renderableList = new GLPbuffer[numValid];
        else
        {
            // Null out what is left of the array of nodes for GC
            for(int i = numValid; i < renderableList.length; i++)
                renderableList[i] = null;
        }

        numRenderables = numValid;

        for(int i = numRenderables; --i >= 0 ; )
        {
            GraphicsInstructions ri = commands[i];

            // Is this the main canvas or one of the offscreen drawables?
            if(ri.pbuffer != null)
            {
                PBufferTextureSource tex = ri.pbuffer;
                GLPbuffer buf = (GLPbuffer)glDrawableMap.get(tex);

                RenderingProcessor rp = null;

                if(buf == null)
                {
                    // If the parent scene is null, then the parent is the
                    // main canvas. Create the offscreen buffer from the parent
                    // canvas first.
                    GLContext parent_ctx;

                    if(ri.parentSource == null)
                        parent_ctx = canvasContext;
                    else
                    {
                        GLAutoDrawable drw =
                            (GLAutoDrawable)glDrawableMap.get(ri.parentSource);
                        parent_ctx = drw.getContext();
                    }

                    if(!canCreatePBuffers)
                        continue;

                    GLCapabilities caps = tex.getGLSetup();
                    int width = tex.getWidth();
                    int height = tex.getHeight();

                    GLDrawableFactory fac = GLDrawableFactory.getFactory();

                    buf = fac.createGLPbuffer(caps, null, width, height, parent_ctx);
                    rp = new StandardRenderingProcessor(buf.getContext());
                    rp.enableSingleThreaded(singleThreaded);

                    tex.registerBuffer(parent_ctx.getGL(), buf);
                    rendererMap.put(buf, rp);
                    glDrawableMap.put(tex, buf);
                }
                else
                {
                    rp = (RenderingProcessor)rendererMap.get(buf);
                }

                renderableList[i] = buf;

                // do we need to go find the original data specification
                // or use the local values?
                if(ri.copyOf != null)
                {
                    ri = (GraphicsInstructions)ri.copyOf;
                }

                tex.getClearColor(colourTmp);
                rp.setClearColor(colourTmp[0],
                                 colourTmp[1],
                                 colourTmp[2],
                                 colourTmp[3]);

                rp.setDrawableObjects(ri.renderList,
                                      ri.renderOps,
                                      ri.numValid,
                                      ri.renderData);
            }
            else
            {
                canvasRenderer.setDrawableObjects(ri.renderList,
                                                  ri.renderOps,
                                                  ri.numValid,
                                                  ri.renderData);

                renderableList[i] = null;
            }
        }
    }

    /**
     * Swap the buffers now if the surface supports multiple buffer drawing.
     * For surfaces that don't support multiple buffers, this does nothing.
     */
    public void swap()
    {
        if(!terminate)
        {
            // JC:
            // Sometimes we manage to get JOGL into a weird state where it has
            // a blocking point internally with its threading model. At points
            // this will throw an InterruptedException from the internals. We
            // really don't care about that, so we just catch and ignore if we
            // happen to see that. Anything else may be important, so we wrap
            // and re-throw it as a RuntimeException.
            try
            {
                canvasRenderer.swapBuffers();
            }
            catch(GLException e)
            {
                if(!(e.getCause() instanceof InterruptedException))
                    throw e;
            }
        }
    }

    /**
     * Get the surface to VWorld transformation matrix.
     *
     * @param x The X coordinate in the entire surface
     * @param y The Y coordinate in the entire surface
     * @param layer The layer ID to fetch from. Layer 0 is the front-most
     * @param subLayer The ID of the viewport-layer that is needed. If there
     *   are no sub-layers, use 0.
     * @param matrix The matrix to copy into
     *
     * @return Whether the coordinates where on the layer
     */
    public boolean getSurfaceToVWorld(int x,
                                   int y,
                                   int layer,
                                   int subLayer,
                                   Matrix4f matrix)
    {
        return canvasRenderer.getSurfaceToVWorld(x, y, layer, subLayer, matrix);
    }

    /**
     * Convert a pixel location to surface coordinates.
     *
     * @param x The X coordinate
     * @param y The Y coordinate
     * @param layer The layer ID to fetch from. Layer 0 is the front-most
     * @param subLayer The ID of the viewport-layer that is needed. If there
     *   are no sub-layers, use 0.
     * @param position The converted position.  It must be preallocated.
     *
     * @return Whether the coordinates where on the layer
     */
    public boolean getPixelLocationInSurface(int x,
                                          int y,
                                          int layer,
                                          int subLayer,
                                          Point3f position)
    {
        return canvasRenderer.getPixelLocationInSurface(x,
                                                 y,
                                                 layer,
                                                 subLayer,
                                                 position);
    }

    /**
     * Get the Center Eye position in surface coordinates.
     *
     * @param x The X coordinate in the entire surface
     * @param y The Y coordinate in the entire surface
     * @param layer The layer ID to fetch from. Layer 0 is the front-most
     * @param subLayer The ID of the viewport-layer that is needed. If there
     *   are no sub-layers, use 0.
     * @param position The current eye position.  It must be preallocated.
     *
     * @return Whether the coordinates where on the layer
     */
    public boolean getCenterEyeInSurface(int x,
                                      int y,
                                      int layer,
                                      int subLayer,
                                      Point3f position)
    {
        return canvasRenderer.getCenterEyeInSurface(x, y, layer, subLayer, position);
    }

    //---------------------------------------------------------------
    // Methods defined by OutputDevice
    //---------------------------------------------------------------

    /**
     * Register an error reporter with the engine so that any errors generated
     * by the node's internals can be reported in a nice, pretty fashion.
     * Setting a value of null will clear the currently set reporter. If one
     * is already set, the new value replaces the old.
     *
     * @param reporter The instance to use or null
     */
    public void setErrorReporter(ErrorReporter reporter)
    {
        if(reporter == null)
            errorReporter = DefaultErrorReporter.getDefaultReporter();
        else
            errorReporter = reporter;

        canvasRenderer.setErrorReporter(reporter);
    }

    /**
     * An object has been detected as being deleted (no internal or
     * external references to it left) and it needs to be cleaned up. The
     * next oppourtunity this drawable and any sub-drawables get, this object
     * should have it's <code>cleanup()</code> method called and then the
     * reference to it should be discarded.
     *
     * @param obj The instance to be queued for deletion
     */
    public void queueDeletedObject(DeletableRenderable obj)
    {
        canvasRenderer.queueDeletedObject(obj);

        // Take local reference in case the setDrawableObjects decides to
        // update values right now.
        int count = numRenderables;
        GLPbuffer[] surfaces = renderableList;

        if(count > 1)
        {
            for(int i = 0; i < count; i++)
            {
                if(surfaces[i] != null)
                {
                    RenderingProcessor rp =
                        (RenderingProcessor)rendererMap.get(surfaces[i]);
                    rp.queueDeletedObject(obj);
                }
            }
        }
    }

    /**
     * Instruct the surface to draw the collected set of nodes now. The
     * registered view environment is used to draw to this surface. If no
     * view is registered, the surface is cleared and then this call is
     * exited. The drawing surface does not swap the buffers at this point.
     * <p>
     * The return value indicates success or failure in the ability to
     * render this frame. Typically it will indicate failure if the
     * underlying surface has been disposed of, either directly through the
     * calling of the method on this interface, or through an internal check
     * mechanism. If failure is indicated, then check to see if the surface has
     * been disposed of and discontinue rendering if it has.
     *
     * @return true if the drawing succeeded, or false if not
     */
    public boolean draw()
    {
        // tell the draw lock that it's ok to run now, so long as it's not called
        // before the canvas has completed initialisation.
        if(!initComplete)
            if(!initCanvas())
                return false;

        // TODO:
        // Would be nice to alter this so that we can dynamically query for
        // available extensions after the first run has started. Would need some
        // sort of flag here that checks for new queries and executes it when set.

        // Take local reference in case the setDrawableObjects decides to
        // update values right now.
        int count = numRenderables;
        GLPbuffer[] surfaces = renderableList;
        boolean draw_continue = true;

        if(count > 1)
        {
            for(int i = 0; i < count && !terminate && draw_continue; i++)
            {
                if(surfaces[i] != null)
                {
                    RenderingProcessor rp =
                        (RenderingProcessor)rendererMap.get(surfaces[i]);
                    draw_continue = rp.render();
                }
            }
        }

        // Always render the main canvas last.
        if(!terminate && draw_continue)
            draw_continue = canvasRenderer.render();

        return !terminate && draw_continue;
    }

    /**
     * Instruct this surface that you have finished with the resources needed
     * and to dispose all rendering resources.
     */
    public void dispose()
    {
        terminate = true;

        canvasRenderer.halt();

        for(int i = 0; i < numRenderables; i++)
        {
            if(renderableList[i] != null)
            {
                RenderingProcessor rp =
                    (RenderingProcessor)rendererMap.get(renderableList[i]);
                rp.halt();
            }
        }
    }

    /**
     * Check to see the disposal state of the surface. Will return true if the
     * {@link #dispose} method has been called or an internal dispose handler
     * has detected the underlying surface is no longer valid to draw to.
     *
     * @return true if the surface is disposed and no longer usable
     */
    public boolean isDisposed()
    {
        return terminate;
    }

    /**
     * Notification that this surface is being drawn to with a single thread.
     * This can be used to optmise internal state handling when needed in a
     * single versus multithreaded environment.
     * <p>
     *
     * This method should never be called by end user code. It is purely for
     * the purposes of the {@link org.j3d.aviatrix3d.management.RenderManager}
     * to inform the device about what state it can expect.
     *
     * @param state true if the device can expect single threaded behaviour
     */
    public void enableSingleThreaded(boolean state)
    {
        singleThreaded = state;

        if(canvasRenderer != null)
            canvasRenderer.enableSingleThreaded(state);

        for(int i = 0; i < numRenderables; i++)
        {
            if(renderableList[i] != null)
            {
                RenderingProcessor rp =
                    (RenderingProcessor)rendererMap.get(renderableList[i]);
                rp.enableSingleThreaded(state);
            }
        }
    }

    /**
     * If the output device is marked as single threaded, this instructs the
     * device that the current rendering thread has exited. Next time the draw
     * method is called, a new rendering context will need to be created for
     * a new incoming thread instance. Also, if any other per-thread resources
     * are around, clean those up now. This is called just before that thread
     * exits.
     */
    public void disposeSingleThreadResources()
    {
        if(canvasRenderer != null)
            canvasRenderer.disposeSingleThreadResources();

        for(int i = 0; i < numRenderables; i++)
        {
            if(renderableList[i] != null)
            {
                RenderingProcessor rp =
                    (RenderingProcessor)rendererMap.get(renderableList[i]);
                rp.disposeSingleThreadResources();
            }
        }
    }

    //---------------------------------------------------------------
    // Local Methods
    //---------------------------------------------------------------

    /**
     * Add an extension string to check for at startup.
     *
     * @param glExtensionString String to check for being valid
     */
    public void checkForExtension(String glExtensionString)
    {
        if(extensionList == null || extensionList.length == numExtensions)
        {
            String[] tmp = new String[numExtensions + 5];

            if(extensionList != null)
                System.arraycopy(extensionList, 0, tmp, 0, numExtensions);
            extensionList = tmp;
        }

        extensionList[numExtensions++] = glExtensionString;
    }

    /**
     * Check to see whether this object is currently shared. If it is shared
     * return true, otherwise return false.
     */
    public boolean isShared()
    {
        return sharedSurface != null;
    }

    /**
     * Used during initialisation of the system for the first time. This is
     * called just after the extension strings have been checked, but before
     * we return back to the main rendering loop. The default implementation is
     * empty.
     * <p>
     * The return value indicates success or failure in the ability to
     * initialise this surface. Typically it will indicate failure if the
     * underlying surface has been disposed of or a failure to find the
     * capabilities needed. The default implementation returns true.
     *
     * @param gl An initialised, current gl context to play with
     * @return true if the initialisation succeeded, or false if not
     */
    public boolean completeCanvasInitialisation(GL gl)
    {
        return true;
    }

    /**
     * Common internal initialisation for the constructors. Derived classes
     * must call this during their constructor otherwise this class will crash
     * in some spectacular ways.
     */
    protected void init()
    {
        GLDrawableFactory fac = GLDrawableFactory.getFactory();
        canCreatePBuffers = fac.canCreateGLPbuffer();

        colourTmp = new float[4];

        initComplete = false;

        renderableList = new GLPbuffer[LIST_START_SIZE];
        numRenderables = 0;

        numExtensions = 0;

        rendererMap = new HashMap();
        glDrawableMap = new HashMap();

        errorReporter = DefaultErrorReporter.getDefaultReporter();
    }

    /**
     * Called by the drawable immediately after the OpenGL context is
     * initialized; the GLContext has already been made current when
     * this method is called. Any surface that overrides the draw method
     * should also call this method if the class variable {@link #initComplete}
     * is currently set to false.
     * <p>
     * The return value indicates success or failure in the ability to
     * initialise this surface. Typically it will indicate failure if the
     * underlying surface has been disposed of or a failure to find the
     * capabilities needed. The default implementation returns true.
     *
     * @return true if the initialisation succeeded, or false if not
     */
    protected boolean initCanvas()
    {
        if(canvasContext == null)
            return false;

        int status = canvasContext.makeCurrent();

        switch(status)
        {
            case GLContext.CONTEXT_NOT_CURRENT:
                // Exit right now as there's nothing left to do.
                errorReporter.errorReport(FAILED_CONTEXT_MSG, null);
                return false;
        }

        GL gl = canvasContext.getGL();

        // Check for extensions:
        for(int i = 0; i < numExtensions; i++)
        {
            if(!gl.isExtensionAvailable(extensionList[i]))
                System.out.println("Extension " + extensionList[i] +
                                   " not available");
        }

        if(completeCanvasInitialisation(gl))
        {
            canvasContext.release();
            initComplete = true;
        }

        return initComplete;
    }

    /**
     * Package local method to fetch the GLContext that this surface has.
     * Allows there to be sharing between different surface types - for example
     * having an elumens surface on one screen and a normal renderer on
     * another.
     *
     * @return The context used by this surface
     */
    GLContext getGLContext()
    {
        return canvasContext;
    }

    /**
     * Get the context object from the shared surface, if there is one set.
     * If there is no surface set then this will return null
     *
     * @return The context from the shared surface or null
     */
    protected GLContext getSharedGLContext()
    {
        return  (sharedSurface != null) ? sharedSurface.getGLContext() : null;
    }
}
